@@ -22,6 +22,8 @@ from .retry import retry_request, validate_response
 from .checkers import get_consistency_checker
 from .credentials import GoogleCredentials
 from . import __version__ as version
+from urllib.parse import quote as quote_urllib
+from urllib.parse import parse_qs, urlsplit
 
 logger = logging.getLogger("gcsfs")
 
@@ -48,18 +50,9 @@ bACLs = {
 }
 DEFAULT_PROJECT = os.environ.get("GCSFS_DEFAULT_PROJECT", "")
 
-GCS_MIN_BLOCK_SIZE = 2 ** 18
-GCS_MAX_BLOCK_SIZE = 2 ** 28
-DEFAULT_BLOCK_SIZE = 5 * 2 ** 20
-
-
-QUOTE_TABLE = str.maketrans(
-    {
-        "%": "%25",
-        "/": "%2F",
-        " ": "%20",
-    }
-)
+GCS_MIN_BLOCK_SIZE = 2**18
+GCS_MAX_BLOCK_SIZE = 2**28
+DEFAULT_BLOCK_SIZE = 5 * 2**20
 
 SUPPORTED_FIXED_KEY_METADATA = {
     "content_encoding": "contentEncoding",
@@ -70,12 +63,10 @@ SUPPORTED_FIXED_KEY_METADATA = {
 }
 
 
-def quote_plus(s):
+def quote(s):
     """
-    Convert some URL elements to be HTTP-safe.
-
-    Not the same as in urllib, because, for instance, parentheses and commas
-    are passed through.
+    Quote characters to be safe for URL paths.
+    Also quotes '/'.
 
     Parameters
     ----------
@@ -85,7 +76,8 @@ def quote_plus(s):
     -------
     corrected URL
     """
-    return s.translate(QUOTE_TABLE)
+    # Encode everything, including slashes
+    return quote_urllib(s, safe="")
 
 
 def norm_path(path):
@@ -94,7 +86,8 @@ def norm_path(path):
 
     Used by petastorm, do not remove.
     """
-    return "/".join(GCSFileSystem.split_path(path))
+    bucket, name, _ = GCSFileSystem._split_path(path)
+    return "/".join((bucket, name))
 
 
 async def _req_to_text(r):
@@ -118,6 +111,32 @@ def _location():
     )
 
 
+def _chunks(lst, n):
+    """
+    Yield evenly-sized chunks from a list.
+
+    Implementation based on https://stackoverflow.com/a/312464.
+    """
+    for i in range(0, len(lst), n):
+        yield lst[i : i + n]
+
+
+def _coalesce_generation(*args):
+    """Helper to coalesce a list of object generations down to one."""
+    generations = set(args)
+    if None in generations:
+        generations.remove(None)
+    if len(generations) > 1:
+        raise ValueError(
+            "Cannot coalesce generations where more than one are defined,"
+            " {}".format(generations)
+        )
+    elif len(generations) == 0:
+        return None
+    else:
+        return generations.pop()
+
+
 class GCSFileSystem(AsyncFileSystem):
     r"""
     Connect to Google Cloud Storage.
@@ -203,12 +222,6 @@ class GCSFileSystem(AsyncFileSystem):
         Cache expiration time in seconds for object metadata cache.
         Set cache_timeout <= 0 for no caching, None for no cache expiration.
     secure_serialize: bool (deprecated)
-    check_connection: bool
-        When token=None, gcsfs will attempt various methods of establishing
-        credentials, falling back to anon. It is possible for a method to
-        find credentials in the system that turn out not to be valid. Setting
-        this parameter to True will ensure that an actual operation is
-        attempted before deciding that credentials are valid.
     requester_pays : bool, or str default False
         Whether to use requester-pays requests. This will include your
         project ID `project` in requests as the `userPorject`, and you'll be
@@ -217,11 +230,18 @@ class GCSFileSystem(AsyncFileSystem):
     session_kwargs: dict
         passed on to aiohttp.ClientSession; can contain, for example,
         proxy settings.
-    endpoin_url: str
+    endpoint_url: str
         If given, use this URL (format protocol://host:port , *without* any
         path part) for communication. If not given, defaults to the value
         of environment variable "STORAGE_EMULATOR_HOST"; if that is not set
         either, will use the standard Google endpoint.
+    default_location: str
+        Default location where buckets are created, like 'US' or 'EUROPE-WEST3'.
+        You can find a list of all available locations here:
+        https://cloud.google.com/storage/docs/locations#available-locations
+    version_aware: bool
+        Whether to support object versioning. If enabled this will require the
+        user to have the necessary permissions for dealing with versioned objects.
     """
 
     scopes = {"read_only", "read_write", "full_control"}
@@ -239,7 +259,7 @@ class GCSFileSystem(AsyncFileSystem):
         consistency="none",
         cache_timeout=None,
         secure_serialize=True,
-        check_connection=False,
+        check_connection=None,
         requests_timeout=None,
         requester_pays=False,
         asynchronous=False,
@@ -247,6 +267,8 @@ class GCSFileSystem(AsyncFileSystem):
         loop=None,
         timeout=None,
         endpoint_url=None,
+        default_location=None,
+        version_aware=False,
         **kwargs,
     ):
         super().__init__(
@@ -270,8 +292,16 @@ class GCSFileSystem(AsyncFileSystem):
         self._session = None
         self._endpoint = endpoint_url
         self.session_kwargs = session_kwargs or {}
+        self.default_location = default_location
+        self.version_aware = version_aware
 
-        self.credentials = GoogleCredentials(project, access, token, check_connection)
+        if check_connection:
+            warnings.warn(
+                "The `check_connection` argument is deprecated and will be removed in a future release.",
+                DeprecationWarning,
+            )
+
+        self.credentials = GoogleCredentials(project, access, token)
 
         if not self.asynchronous:
             self._session = sync(
@@ -327,6 +357,13 @@ class GCSFileSystem(AsyncFileSystem):
         # use of root_marker to make minimum required path, e.g., "/"
         return path or cls.root_marker
 
+    @classmethod
+    def _get_kwargs_from_urls(cls, path):
+        _, _, generation = cls._split_path(path, version_aware=True)
+        if generation is not None:
+            return {"version_aware": True}
+        return {}
+
     def _get_params(self, kwargs):
         params = {k: v for k, v in kwargs.items() if v is not None}
         # needed for requester pays buckets
@@ -352,7 +389,7 @@ class GCSFileSystem(AsyncFileSystem):
             path = self.base + path
 
         if args:
-            path = path.format(*[quote_plus(p) for p in args])
+            path = path.format(*[quote(p) for p in args])
         return path
 
     @retry_request(retries=retries)
@@ -382,7 +419,6 @@ class GCSFileSystem(AsyncFileSystem):
         self, method, path, *args, json_out=False, info_out=False, **kwargs
     ):
         logger.debug(f"{method.upper()}: {path}, {args}, {kwargs.get('headers')}")
-
         status, headers, info, contents = await self._request(
             method, path, *args, **kwargs
         )
@@ -402,8 +438,7 @@ class GCSFileSystem(AsyncFileSystem):
             b["name"] for b in sync(self.loop, self._list_buckets, timeout=self.timeout)
         ]
 
-    @staticmethod
-    def _process_object(bucket, object_metadata):
+    def _process_object(self, bucket, object_metadata):
         """Process object resource into gcsfs object information format.
 
         Process GCS object resource via type casting and attribute updates to
@@ -416,6 +451,9 @@ class GCSFileSystem(AsyncFileSystem):
         result["size"] = int(object_metadata.get("size", 0))
         result["name"] = posixpath.join(bucket, object_metadata["name"])
         result["type"] = "file"
+        if "generation" in object_metadata or "metageneration" in object_metadata:
+            result["generation"] = object_metadata.get("generation")
+            result["metageneration"] = object_metadata.get("metageneration")
 
         return result
 
@@ -429,13 +467,18 @@ class GCSFileSystem(AsyncFileSystem):
 
     async def _get_object(self, path):
         """Return object information at the given path."""
-        bucket, key = self.split_path(path)
+        bucket, key, generation = self.split_path(path)
 
         # Check if parent dir is in listing cache
         listing = self._ls_from_cache(path)
         if listing:
+            name = "/".join((bucket, key))
             for file_details in listing:
-                if file_details["type"] == "file" and file_details["name"] == path:
+                if (
+                    file_details["type"] == "file"
+                    and file_details["name"] == name
+                    and (not generation or file_details.get("generation") == generation)
+                ):
                     return file_details
             else:
                 raise FileNotFoundError(path)
@@ -449,23 +492,33 @@ class GCSFileSystem(AsyncFileSystem):
         # Work around various permission settings. Prefer an object get (storage.objects.get), but
         # fall back to a bucket list + filter to object name (storage.objects.list).
         try:
-            res = await self._call("GET", "b/{}/o/{}", bucket, key, json_out=True)
+            res = await self._call(
+                "GET", "b/{}/o/{}", bucket, key, json_out=True, generation=generation
+            )
         except OSError as e:
             if not str(e).startswith("Forbidden"):
                 raise
             resp = await self._call(
-                "GET", "b/{}/o", bucket, json_out=True, prefix=key, maxResults=1
+                "GET",
+                "b/{}/o",
+                bucket,
+                json_out=True,
+                prefix=key,
+                maxResults=1 if not generation else None,
+                versions="true" if generation else None,
             )
             for item in resp.get("items", []):
-                if item["name"] == key:
+                if item["name"] == key and (
+                    not generation or item.get("generation") == generation
+                ):
                     res = item
                     break
             if res is None:
                 raise FileNotFoundError(path)
         return self._process_object(bucket, res)
 
-    async def _list_objects(self, path, prefix=""):
-        bucket, key = self.split_path(path)
+    async def _list_objects(self, path, prefix="", versions=False):
+        bucket, key, generation = self.split_path(path)
         path = path.rstrip("/")
 
         try:
@@ -481,7 +534,9 @@ class GCSFileSystem(AsyncFileSystem):
             if key:
                 raise
 
-        items, prefixes = await self._do_list_objects(path, prefix=prefix)
+        items, prefixes = await self._do_list_objects(
+            path, prefix=prefix, versions=versions
+        )
 
         pseudodirs = [
             {
@@ -498,15 +553,17 @@ class GCSFileSystem(AsyncFileSystem):
                 return [await self._get_object(path)]
             else:
                 return []
-        out = items + pseudodirs
+        out = pseudodirs + items
         # Don't cache prefixed/partial listings
         if not prefix:
             self.dircache[path] = out
         return out
 
-    async def _do_list_objects(self, path, max_results=None, delimiter="/", prefix=""):
+    async def _do_list_objects(
+        self, path, max_results=None, delimiter="/", prefix="", versions=False
+    ):
         """Object listing for the given {bucket}/{prefix}/ path."""
-        bucket, _path = self.split_path(path)
+        bucket, _path, generation = self.split_path(path)
         _path = "" if not _path else _path.rstrip("/") + "/"
         prefix = f"{_path}{prefix}" or None
 
@@ -520,6 +577,7 @@ class GCSFileSystem(AsyncFileSystem):
             prefix=prefix,
             maxResults=max_results,
             json_out=True,
+            versions="true" if versions or generation else None,
         )
 
         prefixes.extend(page.get("prefixes", []))
@@ -536,6 +594,7 @@ class GCSFileSystem(AsyncFileSystem):
                 maxResults=max_results,
                 pageToken=next_page_token,
                 json_out=True,
+                versions="true" if generation else None,
             )
 
             assert page["kind"] == "storage#objects"
@@ -570,7 +629,8 @@ class GCSFileSystem(AsyncFileSystem):
                 next_page_token = page.get("nextPageToken", None)
 
             buckets = [
-                {"name": i["name"] + "/", "size": 0, "type": "directory"} for i in items
+                {**i, "name": i["name"] + "/", "size": 0, "type": "directory"}
+                for i in items
             ]
             self.dircache[""] = buckets
             return buckets
@@ -597,32 +657,66 @@ class GCSFileSystem(AsyncFileSystem):
                 path = self._parent(path)
 
     async def _mkdir(
-        self, bucket, acl="projectPrivate", default_acl="bucketOwnerFullControl"
+        self,
+        path,
+        acl="projectPrivate",
+        default_acl="bucketOwnerFullControl",
+        location=None,
+        create_parents=True,
+        enable_versioning=False,
+        **kwargs,
     ):
         """
         New bucket
 
+        If path is more than just a bucket, will create bucket if create_parents=True;
+        otherwise is a noop. If create_parents is False and bucket does not exist,
+        will produce FileNotFFoundError.
+
         Parameters
         ----------
-        bucket: str
+        path: str
             bucket name. If contains '/' (i.e., looks like subdir), will
             have no effect because GCS doesn't have real directories.
         acl: string, one of bACLs
             access for the bucket itself
         default_acl: str, one of ACLs
             default ACL for objects created in this bucket
+        location: Optional[str]
+            Location where buckets are created, like 'US' or 'EUROPE-WEST3'.
+            If not provided, defaults to `self.default_location`.
+            You can find a list of all available locations here:
+            https://cloud.google.com/storage/docs/locations#available-locations
+        create_parents: bool
+            If True, creates the bucket in question, if it doesn't already exist
+        enable_versioning: bool
+            If True, creates the bucket in question with object versioning
+            enabled.
         """
+        bucket, object, generation = self.split_path(path)
         if bucket in ["", "/"]:
             raise ValueError("Cannot create root bucket")
-        if "/" in bucket:
+        if "/" in path and create_parents and await self._exists(bucket):
+            # nothing to do
             return
+        if "/" in path and not create_parents:
+            if await self._exists(bucket):
+                return
+            raise FileNotFoundError(bucket)
+
+        json_data = {"name": bucket}
+        location = location or self.default_location
+        if location:
+            json_data["location"] = location
+        if enable_versioning:
+            json_data["versioning"] = {"enabled": True}
         await self._call(
             method="POST",
             path="b",
             predefinedAcl=acl,
             project=self.project,
             predefinedDefaultObjectAcl=default_acl,
-            json={"name": bucket},
+            json=json_data,
             json_out=True,
         )
         self.invalidate_cache(bucket)
@@ -647,20 +741,32 @@ class GCSFileSystem(AsyncFileSystem):
 
     rmdir = sync_wrapper(_rmdir)
 
-    async def _info(self, path, **kwargs):
+    async def _info(self, path, generation=None, **kwargs):
         """File information about this path."""
-        path = self._strip_protocol(path).rstrip("/")
+        path = self._strip_protocol(path)
         if "/" not in path:
-            out = await self._call("GET", f"b/{path}", json_out=True)
-            out.update(size=0, type="directory")
+            try:
+                out = await self._call("GET", f"b/{path}", json_out=True)
+                out.update(size=0, type="directory")
+            except OSError:
+                # GET bucket failed, try ls; will have no metadata
+                exists = await self._ls(path)
+                if exists:
+                    out = {"name": path, "size": 0, "type": "directory"}
+                else:
+                    raise FileNotFoundError(path)
             return out
         # Check directory cache for parent dir
         parent_path = self._parent(path)
         parent_cache = self._ls_from_cache(parent_path)
-        bucket, key = self.split_path(path)
+        bucket, key, path_generation = self.split_path(path)
+        generation = _coalesce_generation(generation, path_generation)
         if parent_cache:
+            name = "/".join((bucket, key))
             for o in parent_cache:
-                if o["name"].rstrip("/") == path:
+                if o["name"].rstrip("/") == name and (
+                    not generation or o.get("generation") == generation
+                ):
                     return o
         if self._ls_from_cache(path):
             # this is a directory
@@ -673,7 +779,10 @@ class GCSFileSystem(AsyncFileSystem):
             }
         # Check exact file path
         try:
-            return await self._get_object(path)
+            exact = await self._get_object(path)
+            # this condition finds a "placeholder" - still need to check if it's a directory
+            if exact["size"] or not exact["name"].endswith("/"):
+                return exact
         except FileNotFoundError:
             pass
         kwargs["detail"] = True  # Force to true for info
@@ -721,14 +830,19 @@ class GCSFileSystem(AsyncFileSystem):
             return sorted([o["name"] for o in out])
 
     def url(self, path):
-        """ Get HTTP URL of the given path """
-        u = "{}/download/storage/v1/b/{}/o/{}?alt=media"
-        bucket, object = self.split_path(path)
-        object = quote_plus(object)
-        return u.format(self._location, bucket, object)
+        """Get HTTP URL of the given path"""
+        u = "{}/download/storage/v1/b/{}/o/{}?alt=media{}"
+        bucket, object, generation = self.split_path(path)
+        object = quote(object)
+        return u.format(
+            self._location,
+            bucket,
+            object,
+            "&generation={}".format(generation) if generation else "",
+        )
 
-    async def _cat_file(self, path, start=None, end=None):
-        """ Simple one-shot get of file data """
+    async def _cat_file(self, path, start=None, end=None, **kwargs):
+        """Simple one-shot get of file data"""
         u2 = self.url(path)
         if start or end:
             head = {"Range": await self._process_limits(path, start, end)}
@@ -754,6 +868,9 @@ class GCSFileSystem(AsyncFileSystem):
     ):
         """Set/delete/add writable metadata attributes
 
+        Note: uses PATCH method (update), leaving unedited keys alone.
+        fake-gcs-server:latest does not seem to support this.
+
         Parameters
         ---------
         content_type: str
@@ -789,7 +906,7 @@ class GCSFileSystem(AsyncFileSystem):
             i_json["contentEncoding"] = content_encoding
         i_json.update(_convert_fixed_key_metadata(fixed_key_metadata))
 
-        bucket, key = self.split_path(path)
+        bucket, key, generation = self.split_path(path)
         o_json = await self._call(
             "PATCH",
             "b/{}/o/{}",
@@ -799,14 +916,13 @@ class GCSFileSystem(AsyncFileSystem):
             json=i_json,
             json_out=True,
         )
-        (await self._info(path))["metadata"] = o_json.get("metadata", {})
         return o_json.get("metadata", {})
 
     setxattrs = sync_wrapper(_setxattrs)
 
     async def _merge(self, path, paths, acl=None):
         """Concatenate objects within a single bucket"""
-        bucket, key = self.split_path(path)
+        bucket, key, generation = self.split_path(path)
         source = [{"name": self.split_path(p)[1]} for p in paths]
         await self._call(
             "POST",
@@ -826,8 +942,10 @@ class GCSFileSystem(AsyncFileSystem):
 
     async def _cp_file(self, path1, path2, acl=None, **kwargs):
         """Duplicate remote file"""
-        b1, k1 = self.split_path(path1)
-        b2, k2 = self.split_path(path2)
+        b1, k1, g1 = self.split_path(path1)
+        b2, k2, g2 = self.split_path(path2)
+        if g2:
+            raise ValueError("Cannot write to specific object generation")
         out = await self._call(
             "POST",
             "b/{}/o/{}/rewriteTo/b/{}/o/{}",
@@ -838,6 +956,7 @@ class GCSFileSystem(AsyncFileSystem):
             headers={"Content-Type": "application/json"},
             destinationPredefinedAcl=acl,
             json_out=True,
+            sourceGeneration=g1,
         )
         while out["done"] is not True:
             out = await self._call(
@@ -851,12 +970,13 @@ class GCSFileSystem(AsyncFileSystem):
                 rewriteToken=out["rewriteToken"],
                 destinationPredefinedAcl=acl,
                 json_out=True,
+                sourceGeneration=g1,
             )
 
     async def _rm_file(self, path, **kwargs):
-        bucket, key = self.split_path(path)
+        bucket, key, generation = self.split_path(path)
         if key:
-            await self._call("DELETE", "b/{}/o/{}", bucket, key)
+            await self._call("DELETE", "b/{}/o/{}", bucket, key, generation=generation)
             self.invalidate_cache(posixpath.dirname(self._strip_protocol(path)))
         else:
             await self._rmdir(path)
@@ -867,41 +987,51 @@ class GCSFileSystem(AsyncFileSystem):
             "Content-Type: application/http\n"
             "Content-Transfer-Encoding: binary\n"
             "Content-ID: <b29c5de2-0db4-490b-b421-6a51b598bd11+{i}>"
-            "\n\nDELETE /storage/v1/b/{bucket}/o/{key} HTTP/1.1\n"
+            "\n\nDELETE /storage/v1/b/{bucket}/o/{key}{query} HTTP/1.1\n"
             "Content-Type: application/json\n"
             "accept: application/json\ncontent-length: 0\n"
         )
-        body = "".join(
-            [
-                template.format(
-                    i=i + 1,
-                    bucket=p.split("/", 1)[0],
-                    key=quote_plus(p.split("/", 1)[1]),
+        errors = []
+        # Splitting requests into 100 chunk batches
+        # See https://cloud.google.com/storage/docs/batch
+        for chunk in _chunks(paths, 100):
+            parts = []
+            for i, p in enumerate(chunk):
+                bucket, key, generation = self.split_path(p)
+                query = f"?generation={generation}" if generation else ""
+                parts.append(
+                    template.format(
+                        i=i + 1,
+                        bucket=quote(bucket),
+                        key=quote(key),
+                        query=query,
+                    )
                 )
-                for i, p in enumerate(paths)
-            ]
-        )
-        headers, content = await self._call(
-            "POST",
-            f"{self._location}/batch/storage/v1",
-            headers={
-                "Content-Type": 'multipart/mixed; boundary="=========='
-                '=====7330845974216740156=="'
-            },
-            data=body + "\n--===============7330845974216740156==--",
-        )
+            body = "".join(parts)
+            headers, content = await self._call(
+                "POST",
+                f"{self._location}/batch/storage/v1",
+                headers={
+                    "Content-Type": 'multipart/mixed; boundary="=========='
+                    '=====7330845974216740156=="'
+                },
+                data=body + "\n--===============7330845974216740156==--",
+            )
 
-        boundary = headers["Content-Type"].split("=", 1)[1]
-        parents = [self._parent(p) for p in paths]
-        [self.invalidate_cache(parent) for parent in parents + list(paths)]
-        txt = content.decode()
-        if any(
-            not ("200 OK" in c or "204 No Content" in c)
-            for c in txt.split(boundary)[1:-1]
-        ):
-            pattern = '"message": "([^"]+)"'
-            out = set(re.findall(pattern, txt))
-            raise OSError(out)
+            boundary = headers["Content-Type"].split("=", 1)[1]
+            parents = [self._parent(p) for p in paths]
+            [self.invalidate_cache(parent) for parent in parents + list(paths)]
+            txt = content.decode()
+            if any(
+                not ("200 OK" in c or "204 No Content" in c)
+                for c in txt.split(boundary)[1:-1]
+            ):
+                pattern = '"message": "([^"]+)"'
+                out = set(re.findall(pattern, txt))
+                errors.extend(out)
+
+        if errors:
+            raise OSError(errors)
 
     @property
     def on_google(self):
@@ -948,15 +1078,15 @@ class GCSFileSystem(AsyncFileSystem):
         metadata=None,
         consistency=None,
         content_type="application/octet-stream",
-        chunksize=50 * 2 ** 20,
+        chunksize=50 * 2**20,
     ):
         # enforce blocksize should be a multiple of 2**18
         consistency = consistency or self.consistency
-        bucket, key = self.split_path(path)
+        bucket, key, generation = self.split_path(path)
         size = len(data)
         out = None
-        if size < 5 * 2 ** 20:
-            return await simple_upload(
+        if size < 5 * 2**20:
+            location = await simple_upload(
                 self, bucket, key, data, metadata, consistency, content_type
             )
         else:
@@ -967,10 +1097,12 @@ class GCSFileSystem(AsyncFileSystem):
                     self, location, bit, offset, size, content_type
                 )
 
-        checker = get_consistency_checker(consistency)
-        checker.update(data)
-        checker.validate_json_response(out)
+            checker = get_consistency_checker(consistency)
+            checker.update(data)
+            checker.validate_json_response(out)
+
         self.invalidate_cache(self._parent(path))
+        return location
 
     async def _put_file(
         self,
@@ -979,7 +1111,7 @@ class GCSFileSystem(AsyncFileSystem):
         metadata=None,
         consistency=None,
         content_type="application/octet-stream",
-        chunksize=50 * 2 ** 20,
+        chunksize=50 * 2**20,
         callback=None,
         **kwargs,
     ):
@@ -989,13 +1121,15 @@ class GCSFileSystem(AsyncFileSystem):
         callback = callback or NoOpCallback()
         consistency = consistency or self.consistency
         checker = get_consistency_checker(consistency)
-        bucket, key = self.split_path(rpath)
+        bucket, key, generation = self.split_path(rpath)
+        if generation:
+            raise ValueError("Cannot write to specific object generation")
         with open(lpath, "rb") as f0:
             size = f0.seek(0, 2)
             f0.seek(0)
             callback.set_size(size)
 
-            if size < 5 * 2 ** 20:
+            if size < 5 * 2**20:
                 await simple_upload(
                     self,
                     bucket,
@@ -1006,7 +1140,7 @@ class GCSFileSystem(AsyncFileSystem):
                     content_type=content_type,
                 )
                 callback.absolute_update(size)
-                return None
+
             else:
                 location = await initiate_upload(
                     self, bucket, key, content_type, metadata
@@ -1023,7 +1157,8 @@ class GCSFileSystem(AsyncFileSystem):
                     callback.absolute_update(offset)
                     checker.update(bit)
 
-            checker.validate_json_response(out)
+                checker.validate_json_response(out)
+
             self.invalidate_cache(self._parent(rpath))
 
     async def _isdir(self, path):
@@ -1032,57 +1167,62 @@ class GCSFileSystem(AsyncFileSystem):
         except IOError:
             return False
 
-    async def _find(self, path, withdirs=False, detail=False, prefix="", **kwargs):
+    async def _find(
+        self, path, withdirs=False, detail=False, prefix="", versions=False, **kwargs
+    ):
         path = self._strip_protocol(path)
-        bucket, key = self.split_path(path)
-        out, _ = await self._do_list_objects(
-            path,
-            delimiter=None,
-            prefix=prefix,
+        bucket, key, generation = self.split_path(path)
+
+        if prefix:
+            _path = "" if not key else key.rstrip("/") + "/"
+            _prefix = f"{_path}{prefix}"
+        else:
+            _prefix = key
+
+        objects, _ = await self._do_list_objects(
+            bucket, delimiter="", prefix=_prefix, versions=versions
         )
-        if not prefix and not out and key:
-            try:
-                out = [
-                    await self._get_object(
-                        path,
-                    )
-                ]
-            except FileNotFoundError:
-                out = []
-        dirs = []
-        sdirs = set()
+
+        dirs = {}
         cache_entries = {}
-        for o in out:
-            par = o["name"]
-            while par:
-                par = self._parent(par)
-                if par not in sdirs:
-                    if len(par) < len(path):
-                        break
-                    sdirs.add(par)
-                    dirs.append(
-                        {
-                            "Key": self.split_path(par)[1],
-                            "Size": 0,
-                            "name": par,
-                            "StorageClass": "DIRECTORY",
-                            "type": "directory",
-                            "size": 0,
-                        }
-                    )
-                # Don't cache "folder-like" objects (ex: "Create Folder" in GCS console) to prevent
-                # masking subfiles in subsequent requests.
-                if not o["name"].endswith("/"):
-                    cache_entries.setdefault(par, []).append(o)
+
+        for obj in objects:
+            parent = self._parent(obj["name"])
+            previous = obj
+
+            while parent:
+                dir_key = self.split_path(parent)[1]
+                if not dir_key or len(parent) < len(path):
+                    break
+
+                dirs[parent] = {
+                    "Key": dir_key,
+                    "Size": 0,
+                    "name": parent,
+                    "StorageClass": "DIRECTORY",
+                    "type": "directory",
+                    "size": 0,
+                }
+
+                cache_entries.setdefault(parent, []).append(previous)
+
+                previous = dirs[parent]
+                parent = self._parent(parent)
+
         if not prefix:
             self.dircache.update(cache_entries)
 
         if withdirs:
-            out = sorted(out + dirs, key=lambda x: x["name"])
+            objects = sorted(objects + list(dirs.values()), key=lambda x: x["name"])
 
         if detail:
-            return {o["name"]: o for o in out}
-        return [o["name"] for o in out]
+            if versions:
+                return {f"{o['name']}#{o['generation']}": o for o in objects}
+            return {o["name"]: o for o in objects}
+
+        if versions:
+            return [f"{o['name']}#{o['generation']}" for o in objects]
+        return [o["name"] for o in objects]
 
     @retry_request(retries=retries)
     async def _get_file_request(
@@ -1136,6 +1276,7 @@ class GCSFileSystem(AsyncFileSystem):
         metadata=None,
         autocommit=True,
         fixed_key_metadata=None,
+        generation=None,
         **kwargs,
     ):
         """
@@ -1162,7 +1303,7 @@ class GCSFileSystem(AsyncFileSystem):
         )
 
     @classmethod
-    def split_path(cls, path):
+    def _split_path(cls, path, version_aware=False):
         """
         Normalise GCS path string into bucket and key.
 
@@ -1170,17 +1311,47 @@ class GCSFileSystem(AsyncFileSystem):
         ----------
         path : string
             Input path, like `gcs://mybucket/path/to/file`.
-            Path is of the form: '[gs|gcs://]bucket[/key]'
+            Path is of the form: '[gs|gcs://]bucket[/key][?querystring][#fragment]'
+
+        GCS allows object generation (object version) to be specified in either
+        the URL fragment or the `generation` query parameter. When provided,
+        the fragment will take priority over the `generation` query paramenter.
 
         Returns
         -------
-            (bucket, key) tuple
+            (bucket, key, generation) tuple
         """
         path = cls._strip_protocol(path).lstrip("/")
         if "/" not in path:
-            return path, ""
-        else:
-            return path.split("/", 1)
+            return path, "", None
+        bucket, keypart = path.split("/", 1)
+        key = keypart
+        generation = None
+        if version_aware:
+            parts = urlsplit(keypart)
+            try:
+                if parts.fragment:
+                    generation = parts.fragment
+                elif parts.query:
+                    parsed = parse_qs(parts.query)
+                    if "generation" in parsed:
+                        generation = parsed["generation"][0]
+                # Sanity check whether this could be a valid generation ID. If
+                # it is not, assume that # or ? characters are supposed to be
+                # part of the object name.
+                if generation is not None:
+                    int(generation)
+                    key = parts.path
+            except ValueError:
+                generation = None
+        return (
+            bucket,
+            key,
+            generation,
+        )
+
+    def split_path(self, path):
+        return self._split_path(path, version_aware=self.version_aware)
 
     def sign(self, path, expiration=100, **kwargs):
         """Create a signed URL representing the given path.
@@ -1205,7 +1376,7 @@ class GCSFileSystem(AsyncFileSystem):
         )
         bucket = client.bucket(bucket)
         blob = bucket.blob(key)
-        return blob.generate_signed_url(expiration=expiration)
+        return blob.generate_signed_url(expiration=expiration, **kwargs)
 
 
 GoogleCredentials.load_tokens()
@@ -1227,6 +1398,7 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
         content_type=None,
         timeout=None,
         fixed_key_metadata=None,
+        generation=None,
         **kwargs,
     ):
         """
@@ -1266,7 +1438,13 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
             https://cloud.google.com/storage/docs/metadata#mutable
         timeout: int
             Timeout seconds for the asynchronous callback.
+        generation: str
+            Object generation.
         """
+        bucket, key, path_generation = gcsfs.split_path(path)
+        if not key:
+            raise OSError("Attempt to open a bucket")
+        self.generation = _coalesce_generation(generation, path_generation)
         super().__init__(
             gcsfs,
             path,
@@ -1277,9 +1455,6 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
             cache_options=cache_options,
             **kwargs,
         )
-        bucket, key = self.fs.split_path(path)
-        if not key:
-            raise OSError("Attempt to open a bucket")
         self.gcsfs = gcsfs
         self.bucket = bucket
         self.key = key
@@ -1303,12 +1478,18 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
                 self.blocksize = GCS_MIN_BLOCK_SIZE
             self.location = None
 
+    @property
+    def details(self):
+        if self._details is None:
+            self._details = self.fs.info(self.path, generation=self.generation)
+        return self._details
+
     def info(self):
-        """ File information about this path """
+        """File information about this path"""
         return self.details
 
     def url(self):
-        """ HTTP link to this file's data """
+        """HTTP link to this file's data"""
         return self.fs.url(self.path)
 
     def _upload_chunk(self, final=False):
@@ -1387,7 +1568,7 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
         self._upload_chunk(final=True)
 
     def _initiate_upload(self):
-        """ Create multi-upload """
+        """Create multi-upload"""
         self.location = sync(
             self.gcsfs.loop,
             initiate_upload,
@@ -1410,7 +1591,7 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
         uid = re.findall("upload_id=([^&=?]+)", self.location)
         self.gcsfs.call(
             "DELETE",
-            f"{self.fs._location}/upload/storage/v1/b/{quote_plus(self.bucket)}/o",
+            f"{self.fs._location}/upload/storage/v1/b/{quote(self.bucket)}/o",
             params={"uploadType": "resumable", "upload_id": uid},
             json_out=True,
         )
@@ -1506,7 +1687,7 @@ async def initiate_upload(
     j.update(_convert_fixed_key_metadata(fixed_key_metadata))
     headers, _ = await fs._call(
         method="POST",
-        path=f"{fs._location}/upload/storage/v1/b/{quote_plus(bucket)}/o",
+        path=f"{fs._location}/upload/storage/v1/b/{quote(bucket)}/o",
         uploadType="resumable",
         json=j,
         headers={"X-Upload-Content-Type": content_type},
@@ -1529,7 +1710,7 @@ async def simple_upload(
     fixed_key_metadata=None,
 ):
     checker = get_consistency_checker(consistency)
-    path = f"{fs._location}/upload/storage/v1/b/{quote_plus(bucket)}/o"
+    path = f"{fs._location}/upload/storage/v1/b/{quote(bucket)}/o"
     metadata = {"name": key}
     if metadatain is not None:
         metadata["metadata"] = metadatain
