@@ -66,7 +66,7 @@ def is_retriable(exception):
     return isinstance(exception, RETRIABLE_EXCEPTIONS)
 
 
-def validate_response(status, content, path):
+def validate_response(status, content, path, args=None):
     """
     Check the requests object r, raise error if it's not ok.
 
@@ -76,6 +76,10 @@ def validate_response(status, content, path):
     path: associated URL path, for error messages
     """
     if status >= 400:
+        if args:
+            from .core import quote
+
+            path = path.format(*[quote(p) for p in args])
         if status == 404:
             raise FileNotFoundError(path)
 
@@ -127,7 +131,13 @@ async def retry_request(func, retries=6, *args, **kwargs):
                 )
                 raise ValueError(msg) from e
             # Special test for 404 to avoid retrying the request
-            if isinstance(e, aiohttp.client_exceptions.ClientError) and e.code == 404:
+            if (
+                isinstance(e, aiohttp.client_exceptions.ClientResponseError)
+                and e.status == 404
+            ):
+                logger.debug("Request returned 404, no retries.")
+                raise e
+            if isinstance(e, HttpError) and e.code == 404:
                 logger.debug("Request returned 404, no retries.")
                 raise e
             if retry == retries - 1:
