@@ -5,21 +5,25 @@ from builtins import FileNotFoundError
 from itertools import chain
 from unittest import mock
 from urllib.parse import urlparse, parse_qs, unquote
+from uuid import uuid4
+
 import pytest
 import requests
 
 from fsspec.utils import seek_delimiter
+from fsspec.asyn import sync
 
 from gcsfs.tests.settings import TEST_BUCKET, TEST_PROJECT, TEST_REQUESTER_PAYS_BUCKET
 from gcsfs.tests.conftest import (
-    files,
-    csv_files,
-    text_files,
     a,
+    allfiles,
     b,
+    csv_files,
+    files,
+    text_files,
 )
 from gcsfs.tests.utils import tempdir, tmpfile
-from gcsfs.core import GCSFileSystem, quote_plus
+from gcsfs.core import GCSFileSystem, quote
 from gcsfs.credentials import GoogleCredentials
 import gcsfs.checkers
 from gcsfs import __version__ as version
@@ -86,16 +90,16 @@ def test_large_upload(gcs):
 
 def test_multi_upload(gcs):
     fn = TEST_BUCKET + "/test"
-    d = b"01234567" * 2 ** 15
+    d = b"01234567" * 2**15
 
     # something to write on close
-    with gcs.open(fn, "wb", content_type="text/plain", block_size=2 ** 18) as f:
+    with gcs.open(fn, "wb", content_type="text/plain", block_size=2**18) as f:
         f.write(d)
         f.write(b"xx")
     assert gcs.cat(fn) == d + b"xx"
     assert gcs.info(fn)["contentType"] == "text/plain"
     # empty buffer on close
-    with gcs.open(fn, "wb", content_type="text/plain", block_size=2 ** 19) as f:
+    with gcs.open(fn, "wb", content_type="text/plain", block_size=2**19) as f:
         f.write(d)
         f.write(b"xx")
         f.write(d)
@@ -103,16 +107,16 @@ def test_multi_upload(gcs):
     assert gcs.info(fn)["contentType"] == "text/plain"
 
     fn = TEST_BUCKET + "/test"
-    d = b"01234567" * 2 ** 15
+    d = b"01234567" * 2**15
 
     # something to write on close
-    with gcs.open(fn, "wb", block_size=2 ** 18) as f:
+    with gcs.open(fn, "wb", block_size=2**18) as f:
         f.write(d)
         f.write(b"xx")
     assert gcs.cat(fn) == d + b"xx"
     assert gcs.info(fn)["contentType"] == "application/octet-stream"
     # empty buffer on close
-    with gcs.open(fn, "wb", block_size=2 ** 19) as f:
+    with gcs.open(fn, "wb", block_size=2**19) as f:
         f.write(d)
         f.write(b"xx")
         f.write(d)
@@ -200,6 +204,22 @@ def test_rm_recursive(gcs):
     assert not gcs.exists(TEST_BUCKET + files[-1])
 
 
+def test_rm_chunked_batch(gcs):
+    files = [f"{TEST_BUCKET}/t{i}" for i in range(303)]
+    for fn in files:
+        gcs.touch(fn)
+
+    files_created = gcs.find(TEST_BUCKET)
+    for fn in files:
+        assert fn in files_created
+
+    gcs.rm(files)
+
+    files_removed = gcs.find(TEST_BUCKET)
+    for fn in files:
+        assert fn not in files_removed
+
+
 def test_file_access(gcs):
     fn = TEST_BUCKET + "/nested/file1"
     data = b"hello\n"
@@ -286,7 +306,7 @@ def test_url(gcs):
     fn = TEST_BUCKET + "/nested/file1"
     url = gcs.url(fn)
     assert "http" in url
-    assert quote_plus("nested/file1") in url
+    assert quote("nested/file1") in url
     with gcs.open(fn) as f:
         assert "http" in f.url()
 
@@ -456,6 +476,36 @@ def test_get_put_file_in_dir(protocol, gcs):
         assert gcs.cat(protocol + TEST_BUCKET + "/temp_dir/accounts.1.json") == data1
 
 
+def test_special_characters_filename(gcs: GCSFileSystem):
+    special_filename = """'!"`#$%&'()+,-.<=>?@[]^_{}~/'"""
+    full_path = TEST_BUCKET + "/" + special_filename
+    gcs.touch(full_path)
+    info = gcs.info(full_path)
+    assert info["name"] == full_path
+    # Normal cat currently doesn't work with special characters,
+    # because it invokes expand_path (and in turn glob) without escaping the characters.
+    # This would need to be fixed in fsspec.
+    assert gcs.cat_file(full_path) == b""
+
+
+def test_slash_filename(gcs: GCSFileSystem):
+    slash_filename = """abc/def"""
+    full_path = TEST_BUCKET + "/" + slash_filename
+    gcs.touch(full_path)
+    info = gcs.info(full_path)
+    assert info["name"] == full_path
+    assert gcs.cat_file(full_path) == b""
+
+
+def test_hash_filename(gcs: GCSFileSystem):
+    slash_filename = """a#b#c"""
+    full_path = TEST_BUCKET + "/" + slash_filename
+    gcs.touch(full_path)
+    info = gcs.info(full_path)
+    assert info["name"] == full_path
+    assert gcs.cat_file(full_path) == b""
+
+
 def test_errors(gcs):
     with pytest.raises((IOError, OSError)):
         gcs.open(TEST_BUCKET + "/tmp/test/shfoshf", "rb")
@@ -584,7 +634,7 @@ def text_mode(gcs):
 
 
 def test_write_blocks(gcs):
-    with gcs.open(TEST_BUCKET + "/temp", "wb", block_size=2 ** 18) as f:
+    with gcs.open(TEST_BUCKET + "/temp", "wb", block_size=2**18) as f:
         f.write(b"a" * 100000)
         assert f.buffer.tell() == 100000
         assert not (f.offset)
@@ -597,11 +647,11 @@ def test_write_blocks(gcs):
 def test_write_blocks2(gcs):
     if not gcs.on_google:
         pytest.skip("emulator always accepts whole request")
-    with gcs.open(TEST_BUCKET + "/temp1", "wb", block_size=2 ** 18) as f:
-        f.write(b"a" * (2 ** 18 + 1))
+    with gcs.open(TEST_BUCKET + "/temp1", "wb", block_size=2**18) as f:
+        f.write(b"a" * (2**18 + 1))
         # leftover bytes: GCS accepts blocks in multiples of 2**18 bytes
         assert f.buffer.tell() == 1
-    assert gcs.info(TEST_BUCKET + "/temp1")["size"] == 2 ** 18 + 1
+    assert gcs.info(TEST_BUCKET + "/temp1")["size"] == 2**18 + 1
 
 
 def test_readline(gcs):
@@ -647,16 +697,16 @@ def test_readline_empty(gcs):
 
 
 def test_readline_blocksize(gcs):
-    data = b"ab\n" + b"a" * (2 ** 18) + b"\nab"
+    data = b"ab\n" + b"a" * (2**18) + b"\nab"
     with gcs.open(a, "wb") as f:
         f.write(data)
-    with gcs.open(a, "rb", block_size=2 ** 18) as f:
+    with gcs.open(a, "rb", block_size=2**18) as f:
         result = f.readline()
         expected = b"ab\n"
         assert result == expected
 
         result = f.readline()
-        expected = b"a" * (2 ** 18) + b"\n"
+        expected = b"a" * (2**18) + b"\n"
         assert result == expected
 
         result = f.readline()
@@ -769,7 +819,7 @@ def test_bigger_than_block_read(gcs):
 
 def test_current(gcs):
     assert GCSFileSystem.current() is gcs
-    gcs2 = GCSFileSystem(endpoint_url=gcs._endpoint)
+    gcs2 = GCSFileSystem(endpoint_url=gcs._endpoint, default_location=None)
     assert gcs2.session is gcs.session
 
 
@@ -787,6 +837,9 @@ def test_array(gcs):
 
 
 def test_attrs(gcs):
+    if not gcs.on_google:
+        # https://github.com/fsspec/gcsfs/pull/479
+        pytest.skip("fake-gcs-server:latest only supports PUT for metadata, not PATCH")
     gcs.touch(a)
     assert "metadata" not in gcs.info(a)
     with pytest.raises(KeyError):
@@ -914,6 +967,38 @@ def test_placeholder_dir_cache_validity(gcs):
     gcs.info(f"gs://{TEST_BUCKET}/b")
 
 
+def test_pipe_small_cache_validity(gcs):
+    folder = f"{TEST_BUCKET}/{str(uuid4())}"
+
+    gcs.pipe(f"gs://{folder}/a/file.txt", b"")
+
+    assert gcs.ls(f"gs://{folder}") == [f"{folder}/a"]
+
+    gcs.pipe(f"gs://{folder}/b/file.txt", b"")
+
+    ls_res = gcs.ls(f"gs://{folder}")
+    assert len(ls_res) == 2
+    assert f"{folder}/b" in ls_res
+
+
+def test_put_small_cache_validity(gcs):
+    folder = f"{TEST_BUCKET}/{str(uuid4())}"
+
+    gcs.pipe(f"gs://{folder}/a/file.txt", b"")
+
+    assert gcs.ls(f"gs://{folder}") == [f"{folder}/a"]
+
+    with tmpfile() as fn:
+        with open(fn, "w") as f:
+            f.write("")
+
+        gcs.put(fn, f"gs://{folder}/b/file.txt", b"")
+
+    ls_res = gcs.ls(f"gs://{folder}")
+    assert len(ls_res) == 2
+    assert f"{folder}/b" in ls_res
+
+
 def test_pseudo_dir_find(gcs):
     gcs.rm(f"{TEST_BUCKET}/*", recursive=True)
     gcs.touch(f"{TEST_BUCKET}/a/b/file")
@@ -944,26 +1029,42 @@ def test_zero_cache_timeout(gcs):
         gcs.dircache[f"{TEST_BUCKET}/a"]
 
 
-def test_find_with_prefix_partial_cache(gcs):
+@pytest.mark.parametrize("with_cache", (False, True))
+def test_find_with_prefix_partial_cache(gcs, with_cache):
     base_dir = f"{TEST_BUCKET}/test_find_with_prefix"
     gcs.touch(base_dir + "/test_1")
     gcs.touch(base_dir + "/test_2")
 
-    for with_cache in (True, False):
-        # Test once with cached, and once with no cache
-        gcs.invalidate_cache()
-        if with_cache:
-            gcs.ls(base_dir)
-        precache = dict(gcs.dircache)
-        assert gcs.find(base_dir, prefix="non_existent_") == []
-        assert gcs.find(base_dir, prefix="test_") == [
-            base_dir + "/test_1",
-            base_dir + "/test_2",
-        ]
-        assert dict(gcs.dircache) == precache  # find qwith prefix shouldn't touch cache
-        assert gcs.find(base_dir + "/test_1") == [base_dir + "/test_1"]
-        assert gcs.find(base_dir + "/non_existent") == []
-        assert gcs.find(base_dir + "/non_existent", prefix="more_non_existent") == []
+    gcs.invalidate_cache()
+    if with_cache:
+        gcs.ls(base_dir)
+    precache = dict(gcs.dircache)
+    assert gcs.find(base_dir, prefix="non_existent_") == []
+    assert gcs.find(base_dir, prefix="test_") == [
+        base_dir + "/test_1",
+        base_dir + "/test_2",
+    ]
+    assert dict(gcs.dircache) == precache  # find qwith prefix shouldn't touch cache
+    assert gcs.find(base_dir + "/test_1") == [base_dir + "/test_1"]
+    assert gcs.find(base_dir + "/non_existent") == []
+    assert gcs.find(base_dir + "/non_existent", prefix="more_non_existent") == []
+
+
+def test_find_dircache(gcs):
+    """Running `ls` after find should not corrupt the dir cache"""
+    assert set(gcs.find(TEST_BUCKET)) == {f"{TEST_BUCKET}/{path}" for path in allfiles}
+    assert set(gcs.ls(TEST_BUCKET)) == {
+        f"{TEST_BUCKET}/test",
+        f"{TEST_BUCKET}/nested",
+        f"{TEST_BUCKET}/2014-01-01.csv",
+        f"{TEST_BUCKET}/2014-01-02.csv",
+        f"{TEST_BUCKET}/2014-01-03.csv",
+    }
+    assert set(gcs.ls(f"{TEST_BUCKET}/nested")) == {
+        f"{TEST_BUCKET}/nested/file1",
+        f"{TEST_BUCKET}/nested/file2",
+        f"{TEST_BUCKET}/nested/nested2",
+    }
 
 
 def test_percent_file_name(gcs):
@@ -977,3 +1078,129 @@ def test_percent_file_name(gcs):
     gcs.touch(fn2)
     assert gcs.cat(fn2) != data
     assert set(gcs.ls(parent)) == set([fn, fn2])
+
+
+@pytest.mark.parametrize(
+    "location",
+    [
+        (None),
+        ("US"),
+        ("EUROPE-WEST3"),
+        ("europe-west3"),
+    ],
+)
+def test_bucket_location(gcs_factory, location):
+    gcs = gcs_factory(default_location=location)
+    if not gcs.on_google:
+        pytest.skip("emulator can only create buckets in the 'US-CENTRAL1' location.")
+    bucket_name = str(uuid4())
+    try:
+        gcs.mkdir(bucket_name)
+        bucket = [
+            b
+            for b in sync(gcs.loop, gcs._list_buckets, timeout=gcs.timeout)
+            if b["name"] == bucket_name + "/"
+        ][0]
+        assert bucket["location"] == (location or "US").upper()
+    finally:
+        gcs.rm(bucket_name, recursive=True)
+
+
+def test_bucket_default_location_overwrite(gcs_factory):
+    gcs = gcs_factory(default_location="US")
+    if not gcs.on_google:
+        pytest.skip("emulator can only create buckets in the 'US-CENTRAL1' location.")
+    bucket_name = str(uuid4())
+    try:
+        gcs.mkdir(bucket_name, location="EUROPE-WEST3")
+        bucket = [
+            b
+            for b in sync(gcs.loop, gcs._list_buckets, timeout=gcs.timeout)
+            if b["name"] == bucket_name + "/"
+        ][0]
+        assert bucket["location"] == "EUROPE-WEST3"
+    finally:
+        gcs.rm(bucket_name, recursive=True)
+
+
+def test_dir_marker(gcs):
+    gcs.touch(f"{TEST_BUCKET}/placeholder/")
+    gcs.touch(f"{TEST_BUCKET}/placeholder/inner")
+    out = gcs.find(TEST_BUCKET)
+    assert f"{TEST_BUCKET}/placeholder/" in out
+    gcs.invalidate_cache()
+    out2 = gcs.info(f"{TEST_BUCKET}/placeholder/")
+    out3 = gcs.info(f"{TEST_BUCKET}/placeholder/")
+    assert out2 == out3
+    assert out2["type"] == "directory"
+
+
+def test_mkdir_with_path(gcs):
+
+    with pytest.raises(FileNotFoundError):
+        gcs.mkdir(f"{TEST_BUCKET + 'new'}/path", create_parents=False)
+    assert not gcs.exists(f"{TEST_BUCKET + 'new'}")
+    gcs.mkdir(f"{TEST_BUCKET + 'new'}/path", create_parents=True)
+    assert gcs.exists(f"{TEST_BUCKET + 'new'}")
+
+    # these lines do nothing, but should not fail
+    gcs.mkdir(f"{TEST_BUCKET + 'new'}/path", create_parents=False)
+    gcs.mkdir(f"{TEST_BUCKET + 'new'}/path", create_parents=True)
+
+    gcs.rm(f"{TEST_BUCKET + 'new'}", recursive=True)
+
+
+def test_deep_find_wthdirs(gcs):
+    gcs.touch(f"{TEST_BUCKET}/deep/nested/dir")
+    assert gcs.find(f"{TEST_BUCKET}/deep/nested") == [f"{TEST_BUCKET}/deep/nested/dir"]
+    assert gcs.find(f"{TEST_BUCKET}/deep/nested", withdirs=True) == [
+        f"{TEST_BUCKET}/deep/nested",
+        f"{TEST_BUCKET}/deep/nested/dir",
+    ]
+
+
+def test_info_versioned(gcs_versioned):
+    with gcs_versioned.open(a, "wb") as wo:
+        wo.write(b"v1")
+    v1 = gcs_versioned.info(a)["generation"]
+    assert v1 is not None
+    with gcs_versioned.open(a, "wb") as wo:
+        wo.write(b"v2")
+    v2 = gcs_versioned.info(a)["generation"]
+    assert v2 is not None and v1 != v2
+    assert gcs_versioned.info(f"{a}#{v1}")["generation"] == v1
+    assert gcs_versioned.info(f"{a}?generation={v2}")["generation"] == v2
+
+
+def test_cat_versioned(gcs_versioned):
+    with gcs_versioned.open(a, "wb") as wo:
+        wo.write(b"v1")
+    v1 = gcs_versioned.info(a)["generation"]
+    assert v1 is not None
+    with gcs_versioned.open(a, "wb") as wo:
+        wo.write(b"v2")
+    gcs_versioned.cat(f"{a}#{v1}") == b"v1"
+
+
+def test_cp_versioned(gcs_versioned):
+    with gcs_versioned.open(a, "wb") as wo:
+        wo.write(b"v1")
+    v1 = gcs_versioned.info(a)["generation"]
+    assert v1 is not None
+    with gcs_versioned.open(a, "wb") as wo:
+        wo.write(b"v2")
+    gcs_versioned.cp_file(f"{a}#{v1}", b)
+    assert gcs_versioned.cat(b) == b"v1"
+
+
+def test_find_versioned(gcs_versioned):
+    with gcs_versioned.open(a, "wb") as wo:
+        wo.write(b"v1")
+    v1 = gcs_versioned.info(a)["generation"]
+    with gcs_versioned.open(a, "wb") as wo:
+        wo.write(b"v2")
+    v2 = gcs_versioned.info(a)["generation"]
+    assert {f"{a}#{v1}", f"{a}#{v2}"} == set(gcs_versioned.find(a, versions=True))
+    assert {f"{a}#{v1}", f"{a}#{v2}"} == set(
+        gcs_versioned.find(a, detail=True, versions=True)
+    )
